: utf-8
"""
Process Belgian files.  The goal is for all athletes, competitions, events and results
to end up in OpenTrack and be available for viewing and as JSON.


To set up for local testing, you need to create directory and copy the files down:
In your raceresults environment, execute
    mkdir downloads
    mkdir downloads/atletiek.be
    cd downloads/atletiek.be
    scp rptlab@app1.reportlab.com:/home/rptlab/website/opentrack.run/downloads/atletiek.be/*.xlsx .


7: Kangaroo (1 in database)
8,9: Benjamin
10,11: Pupil
12,13: Miniem
14,15: Cadet
16,17: Scholier
18, 19: Junior
20+ Senior

Ludwig says:
    We use these single-char identification for categories :
    Master          V
    Senior          S
    Junior          J       = U20
    Scholier        L       = U18
    Cadet           K       = U16
    Miniem          M       = U14
    Pupil           P       = U12
    Benjamin        B       = U10
"""

import json, os, sys, xlrd, hashlib, time, textwrap
from pprint import pprint
from datetime import date
from collections import defaultdict
from dateutil.parser import parse as parse_date
from django.conf import settings
from django.core.management.base import BaseCommand

from mongoengine import DoesNotExist, NotUniqueError
from project.main.models import checksum_matches, set_checksum
from project.reference.models import Organisation, Competition, Venue, Role, Person, Affiliation, Country
from project.reference.findathlete import find_athlete
from project.reference.geo import query_lat_lon
from project.reference.agegroups import iaaf_age_group, belgian_category_suffix
from unidecode import unidecode
from reportlab.lib.utils import asBytes, isStr

from project.reference.mongo import Competition as MongoComp, EventSpec, Unit, Competitor
from project.summary.calc import calculate_competition_summary
from docengine.common import xlsx2xls


INPUTDIR = os.path.normpath(os.path.join(settings.ENVIRON_DIR,'..','downloads/atletiek.be'))
TESTDIR = os.path.join(settings.ENVIRON_DIR, 'downloads', 'atletiek.be')

_lat_lon_cache = {}  # do not look up same track 50 times...
def getInputDir():
    d = TESTDIR if os.path.isdir(TESTDIR) else INPUTDIR
    return os.path.relpath(d,settings.ENVIRON_DIR)


def results_suffix(category):
    "Take their column J and return a suffix"
    words = category.split()
    if len(words) == 2:
        cat, gender = words
        prefix = cat[0:3]
        gender = gender[0].upper()
        if prefix in ('Sen', 'Mas'):
            return 'A' + gender
        else:
            return prefix[0] + gender


def season_end_date(comp_date):
    "e.g. return nextyear-10-31 for anything after thisyear-11-01"

    if comp_date.month in [11, 12]:
        yyyy = comp_date.year + 1
    else:
        yyyy = comp_date.year
    return date(yyyy, 10, 31)

class DEFAULTSTR(str):
    pass
class Command(BaseCommand):

    help = "Make athletes from member spreadsheet"
    prefixes = ("Member", "Event", "Results")

    def add_arguments(self,parser):
        for pfx in self.prefixes:
            w = pfx.lower()
            parser.add_argument("--%s-file"%w, dest="%s_file" % w,
                  help="read %s data from file\ndefault=%%(default)s\nspecify empty to skip" % w,
                  default=DEFAULTSTR(os.path.join(getInputDir(),'%s-Opentrack.xlsx' % pfx)),
                  metavar="FILE")

        parser.add_argument("-c", "--comp-num", dest="comp_num_wanted", help="Only process competition with given number")
        parser.add_argument("--force-all-event-lists", dest="force_all_event_lists",
                default=False, action='store_true', help="force changing of all event lists")
        parser.add_argument("--force-event-list", dest="force_event_list",
                default=False, action='store_true', help="force changing of event list for single competition")
        parser.add_argument("--skip", dest="processes_to_skip", help="blank separated list of actions to skip from member event results", default='')
        parser.add_argument("--aws-download", dest="awsDownload", action="store_true", default=False,
                help="specify this flag to download the latest aws files and quit",
                )
        parser.add_argument("--output-id", dest="outputId", action="store", default='',
                help="use this string as a prefix id for stdout/stderr",
                )

    def handle(self, *args, **options):
        outputId = options['outputId']
        if outputId:
            from docengine.tasks.management.commands.run_task import IdentiFile
            sys.stderr = IdentiFile(sys.__stderr__,'!!!!! %s' % outputId)
            sys.stdout = IdentiFile(sys.__stdout__,'+++++ %s' % outputId)
        inputdir = getInputDir()

        verbosity = options['verbosity']

        if options['awsDownload']:
            from project.utils import s3Download
            s3Download(
                    bucketName = settings.AWS_PRIVATE_BUCKET_NAME,
                    objectNames = [
                        'imports/atletiek.be/latest/Event-Opentrack.xlsx',
                        'imports/atletiek.be/latest/Member-Opentrack.xlsx',
                        'imports/atletiek.be/latest/Results-Opentrack.xlsx',
                        'imports/atletiek.be/latest/Results-Opentrack.csv',
                        ],
                    dstDir = inputdir,
                    verbosity = verbosity,
                    )
            return
        comp_num_wanted = options.get('comp_num_wanted', None)

        if verbosity:
            print("Processing default files in", inputdir, "...", end=' ')
            if comp_num_wanted:
                print("for competition", comp_num_wanted)
            else:
                print()

        for pfx in self.prefixes:
            w = pfx.lower()
            k = '%s_file' % w
            filename = options[k] = options[k].strip()
            if not filename or w in options["processes_to_skip"].lower(): continue
            if not os.path.isabs(filename):
                if isinstance(filename,DEFAULTSTR):
                    base = settings.ENVIRON_DIR
                else:
                    base = os.getcwd()
                filename = os.path.normpath(os.path.join(base,filename))
            if not os.path.isfile(os.path.join(inputdir, filename)):
                if verbosity:
                    print("File not found:", filename)
            else:
                if not w.endswith('s'): w += 's'
                func = 'process_'+w
                if verbosity>=3:
                    print("#AR - calling", func)
                # commented the next two lines after unhandled exception on test site
                #if verbosity>2:
                #   print 'globals()[%s](%r,%r)' % (func, filename,verbosity, comp_num)
                globals()[func](filename,verbosity, options)

def process_members(filename, verbosity, options):
    "Federation members"
    clock_start = time.process_time()

    workbook = xlsx2xls.open_workbook(filename)
    datemode = workbook.datemode
    sheet = workbook.sheet_by_index(0)

    start_date = date(2020, 11, 1)
    end_date = date(2021, 10, 31)
    today = date.today()
    stoday = today.strftime('%Y-%m-%d')
    processed = found = changed = newguys = skipped = 0
    missing_clubs = set()
    problem_persons = set()

    if verbosity:
        print("Members sheet has %d rows of person data" % sheet.nrows)

    def makeMap(queryset):
        C = {}
        for r in queryset.select_related('person'):
            C.setdefault(r.person.id,[]).append(r)
        return C

    fed = Organisation.objects.get(country_id="BEL", code="KBAB")
    fedRMap = makeMap(Role.objects.filter(organisation=fed,role_type='REGISTERED',
                            date_from=start_date,date_to=end_date))
    memberRMap = makeMap(Role.objects.filter(role_type='MEMBER',person__nationality_id='BEL',
                            date_from=start_date,date_to=end_date))

    clubmap = {}
    for club in Organisation.objects.filter(country_id="BEL", org_type="CLUB"):
        clubmap[club.code] = club

        aff = Affiliation.objects.filter(parent=fed, child=club).first()
        if not aff:
            aff = Affiliation(parent=fed, child=club)
            aff.save()
            if verbosity:
                print("affiliated club", club.code)

    uniques = set()
    if verbosity:
        print("federation and %d clubs loaded" % len(clubmap))

    countries = Country.objects.all()
    countrymap = {}
    for country in countries:
        countrymap[country.name] = country.alpha3
    for a, b in list(COUNTRYCODES.items()):
        countrymap[a] = b

    known = Role.objects.filter(organisation=fed).select_related("person")
    lifenum2person = {}
    unseen_persons = []

    for r in known:
        p = r.person
        lifenum2person[r.identifier] = p
        unseen_persons.append(p)

    npx = known.count()
    if verbosity:
        print("Pre-loaded %d Belgians" % known.count())

    def process_person(p=None):
        if not p:
            p = Person(
                    first_name=fn,
                    last_name=ln,
                    gender=g,
                    date_of_birth=dob,
                    nationality_id=nat_id,
                    residence_id="BEL",
                    )
            lifenum2person[identifier] = p
        else:
            p.first_name=fn
            p.last_name=ln
            p.gender=g
            p.date_of_birth=dob
            p.nationality_id=nat_id
            p.residence_id="BEL"
        try:
            p.save()
        except:
            print(p)
            raise
        return p

    def process_federation(p, identifier, season_bib, snp, new):
        end_date = date(snp, 10, 31)
        start_date = date(snp - 1, 11, 1)
        r = None
        if not new:
            try:
                r = fedRMap.get(p.id,None)
                if r:
                    r = r[-1]
                else:
                    r = Role.objects.get(person=p,organisation=fed,role_type='REGISTERED', date_to=end_date)
            except Role.DoesNotExist:
                pass
            else:
                if (r.date_from==start_date
                    and r.date_to==end_date
                    and r.identifier==identifier
                    and r.bib==season_bib):
                    return
        if r:
            r.date_from=start_date
            r.date_to=end_date
            r.identifier=identifier
            r.bib=season_bib
            r.notes='data load %s' % stoday
        else:
            r = Role(
                organisation=fed,
                person=p,
                date_from=start_date,
                date_to=end_date,
                role_type='REGISTERED',
                identifier=identifier,
                bib=season_bib,  # needs suffix for age group
                notes='data load %s' % stoday,
                )
            fedRMap.setdefault(p.id,[]).append(r)
        r.save()
        return True #ie we have modified the person


    def person_differs(p):
        return (p.first_name!=fn
             or p.last_name!=ln
             or p.gender!=g
             or p.date_of_birth!=dob
             or p.nationality_id!=nat_id
             or p.residence_id!="BEL")

    def process_club(new):
        try:
            klub = clubmap[club]
        except KeyError:
            missing_clubs.add(club)
        else:
            r = None
            if not new:
                try:
                    r = memberRMap.get(p.id,None)
                    if r:
                        r = [t for t in r if t.organisation==klub]
                        if r:
                            r = r[-1]
                    else:
                        r = Role.objects.get(person=p,role_type="MEMBER",organisation=klub)
                except Role.DoesNotExist:
                    pass
                else:
                    if r and r.date_to==end_date: return

            if r:
                r.date_from = start_date
                r.date_to = end_date
            else:
                r = Role(
                    organisation=klub,
                    person=p,
                    date_from=start_date,
                    date_to=end_date,
                    role_type='MEMBER',
                    )
                memberRMap.setdefault(p.id,[]).append(r)
            r.save()
            return True #ie we have modified the person

    errors = []
    _today = date.today()
    for rownum in range(1, sheet.nrows):  # ignore header
        sl = sheet.row_values(rownum, 0, 11)
        (lifenum, g, fn, ln, sn, snp, rp, cat, club, dob, nation) = sl

        if snp:
            snp = int(snp)

        try:
            lifenum = int(lifenum)
        except:
            print('lifenum=%r cannot be made an integer' % lifenum)
            continue
        if lifenum<=-1:
            skipped += 1
            continue
        if isinstance(sn,(float,int)):
            if int(sn)==sn: sn = int(sn)
            sn = str(sn)

        if dob<0:
            errors.append('row%d: raw dob=%r is bad for %s %s' % (rownum+1,dob, fn, ln))
            continue

        g = "M" if g=="M" else "F"

        dob = xlrd.xldate.xldate_as_datetime(dob, datemode).date()

        #  Season number only unique within an age group, we add a suffix to it.
        age_group = iaaf_age_group(dob, _today, vets=False)
        suffix = belgian_category_suffix(age_group, g)
        season_bib = sn + suffix

        if nation in countrymap:
            nat_id = countrymap[nation]
        else:
            if verbosity>1:
                print('"%s": "",' % nation)
            nat_id = "BEL"

        # look for athlete using identifier
        
        identifier = str(int(lifenum))
        
        p = lifenum2person.get(identifier,None)
        new = p is None
        #print lifenum, g, fn, ln, sn, snp, rp, cat, club, dob, nation, identifier, new
        if not new:
            if p in unseen_persons:
                unseen_persons.remove(p)
            else:
                problem_persons.add(p)
            found += 1
            if person_differs(p):
                if verbosity==3:
                    print(lifenum, end=' ')
                    for _ in (
                            (p.first_name!=fn,p.first_name,fn),
                            (p.last_name!=ln,p.last_name,ln),
                            (p.gender!=g,p.gender,g),
                            (p.date_of_birth!=dob,p.date_of_birth,dob),
                            (p.nationality_id!=nat_id,p.nationality_id,nat_id),
                            (p.residence_id!="BEL",p.residence_id,"BEL"),
                            ):
                        if _[0]: print('%r!=%r' % (_[1],_[2]), end=' ')
                    print()

                changed += 1
                p = process_person(p=p)
        else:
            newguys += 1
            p = process_person(p=None)
        if snp:  # if we know the bib and the year the bib relates to, make a role
            if process_federation(p, identifier, season_bib, snp, new):
                p = Person.objects.get(id=p.id) #refetch the person with the fed link
        process_club(new=new)

        processed = found + newguys
        if verbosity and processed % 1000 == 0:
            print(processed, end=' ')
            if processed%10000==0:
                print()
            else:
                sys.stdout.flush()

    if verbosity:
        if ((processed // 1000) * 1000) % 10000: print()
        print("%d found in database(%d changed) %d are new and %d skipped in %.2f seconds" % (found, changed, newguys, skipped, time.process_time()-clock_start))
        if missing_clubs:
            print("These clubs are unknown:\n%s" % ' '.join(sorted(k for k in missing_clubs)))
        print("%d of the %d pre-existing persons were not seen in the data" % (len(unseen_persons),npx))
        if problem_persons:
            print("These people were identified multiple times by the data\n%s" % '\n'.join((
                '%s %s %s' % (asBytes(p.first_name), asBytes(p.last_name), p.date_of_birth.strftime('%Y-%m-%d'))
                        for p in problem_persons)))
        if errors:
            print("These errors were seen:\n%s" % '\n'.join(errors))

def gen_bytes(*args):
    for x in args:
        if isStr(x):
            yield asBytes(x)    #returns x or x.encode('utf8')
        else:
            yield asBytes(str(x))

def process_events(filename, verbosity, options):
    "Competitions and Events"
    clock_start = time.process_time()
    comp_num_wanted = options['comp_num_wanted']
    force_all_event_lists = options['force_all_event_lists']
    force_event_list = options['force_event_list']
    if verbosity:
        print("process_events:", filename)
    workbook = xlsx2xls.open_workbook(filename)
    datemode = workbook.datemode
    sheet = workbook.sheet_by_index(0)

    # # scan for distinct categories and events
    # cats_seen = set()
    # discs_seen = set()
    # for rownum in xrange(1, sheet.nrows):
    #     discs_seen.add(sheet.cell_value(rownum, 9))
    #     cats_seen.add(sheet.cell_value(rownum, 10))

    # print "distinct disciplines: %d..." % len(discs_seen)
    # print sorted(discs_seen)
    # print "distinct categories: %d..." % len(cats_seen)
    # print sorted(cats_seen)

    evmap = load_event_map()

    # for each competition, we have a number of rows of [event, category] info.
    # gather this in a separate pass, as the main loop can ignore most rows
    # of the workbook, only looking here when a new competition is seen
    evdata = defaultdict(list)
    for rownum in range(1, sheet.nrows):  # ignore header
        txt_event_num = sheet.cell_value(rownum, 3)
        # if txt_event_num.startswith('F'):
        #     continue  # Walloon competition, not interested at this time
        event_num = str(txt_event_num).strip().lower()
        if not event_num: continue
        evt = sheet.cell_value(rownum, 9)
        cat = sheet.cell_value(rownum, 10)
        evdata[event_num].append((evt, cat))

    comps_in_sheet = set()

    # quick pass looking only at rows with new event_id
    for rownum in range(1, sheet.nrows):
        event_num = sheet.cell_value(rownum, 3) #actually the competition number
        event_num = str(event_num).strip().lower()
        if not event_num: continue
        comps_in_sheet.add(event_num)

    comps_in_db = load_competitions()
    skipped_comps = set()
    if verbosity:
        print("Database contains %d competitions" % len(comps_in_db))
    newcomps = 0
    for c in comps_in_sheet:
        if c not in comps_in_db:
            newcomps += 1
    if verbosity:
        print("Events sheet has %d competitions, of which %d are new" % (len(comps_in_sheet), newcomps))

    comps_created = 0
    comps_modified = 0

    prev_event_num = None
    for rownum in range(1, sheet.nrows):  # ignore header  100):  #
        sl = sheet.row_values(rownum, 0, 25)
        event_id = sl[0]
        name = sl[1]
        # if sl[3].startswith('F'):
        #     continue  # Walloon competition, not interested at this time

        event_num = str(sl[3]).strip().lower()

        if event_num == prev_event_num:  # next event in same competition, of no interest
            continue
        else:
            prev_event_num = event_num
            comps_in_sheet.add(event_num)

        start_date = sl[7]
        start_date = xlrd.xldate.xldate_as_datetime(start_date, datemode).date()
        end_date = sl[8]
        end_date = xlrd.xldate.xldate_as_datetime(end_date, datemode).date()
        if end_date<start_date: end_date = start_date
        discipline = sl[9]
        category = sl[10]
        organiser_club_name = sl[14]
        organiser_club_number = sl[15]
        facname, address1, zipcode, city, country = sl[17:22]

        # compute a checksum of relevant content at Competition level
        competition_checksum_text = b','.join(
                            gen_bytes(name.encode('utf8'), event_num, start_date, end_date, # date as text,
                            organiser_club_name, organiser_club_number,
                            facname, address1, zipcode, city, country,
                            ))
        competition_checksum = hashlib.md5(competition_checksum_text).hexdigest()

        unchanged = checksum_matches("BELGIUM_IMPORT", "COMPETITION", event_num, competition_checksum)
        if not unchanged or (comp_num_wanted and event_num==comp_num_wanted):

            # new or different
            set_checksum("BELGIUM_IMPORT", "COMPETITION", event_num, competition_checksum)

            cancelled = 'AFGELAST' in name

            address = "%s, %s %s" % (address1, zipcode, city)
            if address in _lat_lon_cache:
                ll = _lat_lon_cache[address]
            else:
                # remember for next time
                ll = query_lat_lon(address + ' Belgium')
                _lat_lon_cache[address] = ll

            # if it's new, make a basic competition record
            if event_num not in comps_in_db:
                c = Competition(
                    national_id=event_num,
                    full_name=name,
                    short_name=name if len(name) < 50 else '',
                    slug=str(event_num).lower(),
                    country_id="BEL",
                    address=address,
                    date=start_date,
                    finish_date=end_date,
                    year=start_date.year,
                    latitude=ll[0],
                    longitude=ll[1],
                    cancelled=cancelled,
                    )
                try:
                    c.save()
                except:
                    print(f'{c.national_id=!r}\n{c.full_name=!r}\n{c.short_name=!r}\n{c.slug=!r}\n{c.country_id=!r}\n{c.address=!r}\n{c.date=!r}\n{c.finish_date=!r}\n{c.year=!r}\n{c.latitude=!r}\n{c.longitude=!r}\n{c.cancelled=!r}')
                    raise

                comps_in_db[event_num] = c
                comps_created += 1
                if verbosity:
                    print("Created competition %s: %s, %s" % (asBytes(event_num), asBytes(name), asBytes(address)))
            else:
                # it's there, but something changed.
                comps_modified += 1
                c = Competition.objects.get(country_id="BEL", national_id=event_num)
                c.full_name = name
                c.short_name = name[:32]
                c.slug = str(event_num).lower()
                c.address = address
                c.date = start_date
                c.year = start_date.year
                c.finish_date=end_date
                c.cancelled = cancelled
                c.save()
                if verbosity:
                    print("Updated competition %s: %s, %s" % (asBytes(event_num), asBytes(name), asBytes(address)))


        # Now load up the event data, if it changed
        try:
            sql_comp = comps_in_db[event_num]
        except KeyError:
            if event_num not in skipped_comps:
                print("Comp not found - %s" % event_num)
                skipped_comps.add(event_num) # remember so we only warn once
        events_data = evdata[event_num]
        
        if events_data == [('', '')]:  #walloon, comp record only
            mc = sql_comp.mongo_comp  # ensure created
            continue  # next competition we hope

        events_data.sort()
        events_checksum = event_list_checksum(events_data)
        unchanged = checksum_matches("BELGIUM_IMPORT", "EVENTLIST", event_num, events_checksum)
        if unchanged and not (force_all_event_lists or
                (force_event_list and comp_num_wanted and event_num==comp_num_wanted)):
            continue
        else:
            # Dive into Mongo
            # Delete all stored events and results, and start again.
            mc = MongoComp.objects.get(id=sql_comp.id)
            mc.delete_related_objects()
            mc.take_entries = True  # can see grid etc, probably debatable though
            mc.save()
            ot_ev_id = 1
            for (bel_discipline, bel_cat) in events_data:
                # print bel_discipline, bel_cat,
                ot_cat = category_code(bel_cat)
                if ot_cat.endswith('M'):
                    ot_gender = 'M'
                elif ot_cat.endswith('V'):
                    ot_gender = 'F'
                else:
                    ot_gender = 'MF'
                if bel_discipline in evmap:  # might get a new one?
                    ot_evcode = evmap[bel_discipline]
                else:
                    print('  Unknown event "%s", skipping' % asBytes(bel_discipline))
                    continue
                # print "->", ot_evcode, ot_cat
                event = EventSpec(
                    competition=mc,
                    event_id='%02d' % ot_ev_id,
                    event_code=ot_evcode,
                    category=ot_cat,
                    genders=ot_gender,
                    name=(bel_discipline.title() + ' - ' + bel_cat)[0:30]
                    )
                try:
                    event.save()
                except NotUniqueError:
                    print('error saving event %d: ' % ot_ev_id, asBytes(bel_discipline), asBytes(bel_cat), ot_evcode, ot_gender, ot_ev_id)
                    continue
                ot_ev_id += 1
            print("  created %d events for competition %s" % (len(events_data), event_num))
            set_checksum("BELGIUM_IMPORT", "EVENTLIST", event_num, events_checksum)

    print("Created %d new competitions, modified %d in %.2f seconds" % (comps_created, comps_modified, time.process_time()-clock_start))
    return

def _showList(L,sfx,pfx='!!!!! '):
    if not isinstance(L,list):
        L = L.__self__
    if L:
        n = len(L)
        L = ' '.join(L).replace("u'","'")
        L = textwrap.wrap(L,72)
        print(('%sfollowing %d %s\n%s\n' % (pfx,n,sfx,'\n'.join((pfx+_ for _ in L)))))
        L[:] = []

def process_results(filename, verbosity, options):
    comp_num_wanted = options['comp_num_wanted']
    if verbosity:
        print("Processing results file", filename)
        if comp_num_wanted:
            print("   Only for competition number", comp_num_wanted)

    # The results file is a list of statistics, not always ordered by competition.
    # gather all results by competition ID as first pass.  We only need a few facts
    # Takes about 10 seconds on Andy's laptop with results to min-March; will get
    # a lot bigger.
    clock_start = time.process_time()
    workbook = xlsx2xls.open_workbook(filename)
    datemode = workbook.datemode
    sheet = workbook.sheet_by_index(0)

    results_by_competition = {}
    bad_comp_num = [].append
    for rownum in range(1, sheet.nrows):  # ignore header
        comp_num = sheet.cell_value(rownum, 12)
        try:
            comp_num = str(comp_num).strip().lower()
        except:
            bad_comp_num(ascii(comp_num))
            continue
        if comp_num_wanted and (comp_num != comp_num_wanted):
            continue  # make it run faster that hours in development
        discipline = sheet.cell_value(rownum, 1)
        category = sheet.cell_value(rownum, 9)
        athlete_num = sheet.cell_value(rownum, 32)
        perf_time = sheet.cell_value(rownum, 7)
        perf_dist = sheet.cell_value(rownum, 5)
        place = sheet.cell_value(rownum, 17)

        if comp_num not in results_by_competition:
            results_by_competition[comp_num] = []

        results = results_by_competition[comp_num]

        if perf_time:  #00:00:08.95 for 60m
            while perf_time.startswith('00:'):
                perf_time = perf_time[3:]

        if perf_dist:
            try:
                perf_dist = '%0.2f' % float(perf_dist)
            except ValueError:
                print('HJ value:', perf_dist)

        if not athlete_num:
            continue
        results.append((
            discipline,
            category,
            str(int(athlete_num)),
            str(perf_time or perf_dist), # ensure 2dp
            place and int(place) or 0
            ))

    _showList(bad_comp_num,'bad comp nums')
    clock_parsed = time.process_time()
    if verbosity:
        print("parsed %d results in %0.2f seconds" % (sheet.nrows - 1, clock_parsed - clock_start))

    # keep track of how many results we were able to process
    results_processed = 0
    failed_event_lookups = 0
    failed_discipline_lookups = 0
    failed_athlete_lookups = 0
    changed_comps = 0

    evmap = load_event_map()
    # try to go by competition and make a checksum
    comps_in_db = load_competitions()

    athletes = load_season_athletes()
    clock_lsa = time.process_time()
    if verbosity:
        print(("loaded %d athletes in %0.2f seconds" % (len(athletes),clock_lsa-clock_parsed)))
    not_in_db = [].append
    not_changed = [].append

    for comp_num, comp_results in sorted(results_by_competition.items()):
        # if comp_num != '19004': continue

        comp_results.sort() #sort the tuples [(discipline,category,athlete_num,.....),....

        results_chk_text = hashlib.md5(ascii(comp_results).encode('utf8')).hexdigest()
        unchanged = checksum_matches("BELGIUM_IMPORT", "RESULTS", comp_num, results_chk_text)
        if unchanged:
            not_changed(ascii(comp_num))

        if not unchanged or (comp_num_wanted and comp_num==comp_num_wanted): # manually load one competition
            changed_comps += 1

            # now we need to import them.
            # find the competition
            try:
                sc = comps_in_db[comp_num]
            except KeyError:
                not_in_db(ascii(comp_num))
                continue
            mc = MongoComp.objects.get(id=sc.id)
            if not mc.auto_calc_age_groups:
                mc.auto_calc_age_groups = True
                mc.save()

            if verbosity>=3:
                print("%s: importing %d results into %s" % (comp_num, len(comp_results), mc))

            # establish the Belgian year for it # e.g. 201i or 2020
            belgian_year = season_end_date(mc.date).year
            # zap competitors
            Competitor.objects.filter(competition=mc).delete()

            # zap all units which might have results
            evs = mc.get_events()
            units = Unit.objects.filter(competition=mc)
            for unit in units:
                unit.clear()

            # need a mapping of discipline and category to events.  If we miss, never mind.
            evs_by_code_and_cat = {}
            for ev in evs:
                key = (ev.event_code, ev.category)
                evs_by_code_and_cat[key] = ev


            # results should be in event order, keep track of the current unit etc
            current_ot_evcode = None
            current_ot_cat = None
            current_event = None
            current_unit = None
            bibs_seen = set()
            for discipline, category, athlete_num, perf, place in comp_results:
                athlete_cat_num = athlete_num + category_suffix(category)
                key = (belgian_year, athlete_cat_num)
                if key not in bibs_seen:
                    bibs_seen.add(key)
                    #create competitor if known
                    if key in athletes:
                        p = athletes[key]
                        c = Competitor(
                            competition=mc,
                            competitor_id=athlete_cat_num,
                            first_name=p.first_name,
                            last_name=p.last_name,
                            gender=p.gender,
                            date_of_birth=p.date_of_birth,
                            team_id=p.club_text[0:5]
                            )
                        c.save()
                    else:
                        failed_athlete_lookups += 1

                # find the event to load results into
                try:
                    ot_evcode = evmap[discipline]
                except KeyError:
                    failed_discipline_lookups += 1
                    continue
                ot_cat = category_code(category)

                if (ot_evcode != current_ot_evcode) or (ot_cat != current_ot_cat):
                    # new event; save the last one
                    if current_unit:
                        current_unit.status = 'finished'
                        current_unit.results_status = 'official'

                        # it can get saved in chunks, so first time
                        # we gat 1st place, and next time no re-scoring
                        if not current_event.is_height_event():
                            for res in current_unit.results:
                                res.place = None
                        current_unit.calc_results()
                        # carry on despite errors as we want to see all of them
                        try:
                            current_unit.save()
                        except Exception as e:
                            pprint(current_unit.to_dict())
                            print(e.message)
                        finally:
                            current_event.status = 'complete'
                            current_event.save()

                    key2 = (ot_evcode, ot_cat)
                    try:
                        current_event = evs_by_code_and_cat[key2]
                    except KeyError:
                        # see if there is an event for all categories:
                        key3 = (ot_evcode, 'ALL-' + ot_cat[-1])
                        try:
                            current_event = evs_by_code_and_cat[key3]
                        except KeyError:
                            print('No event found for', comp_num, key2)
                            failed_event_lookups += 1
                            continue
                    current_unit = current_event.get_unit(1, 1)
                    if current_event.is_height_event():
                        filler = 'NH'
                    elif current_event.is_length_event():
                        filler = 'NM'
                    else:
                        filler = 'DNF'


                # append the result to the current unit
                current_unit.results.append(current_unit.new_result(
                    place=place,    # ensure cleared so it calculates places
                    bib=athlete_num + category_suffix(category),
                    performance=perf or filler
                    ))
                # un = current_unit
                # print ev, "+", athlete_num, perf, '=', len(un.results)
                results_processed += 1

            # Now save the last unit
            if current_unit:
                current_unit.status = 'finished'
                current_unit.results_status = 'official'
                if not current_event.is_height_event():
                    for res in current_unit.results:
                        res.place = None

                current_unit.calc_results()
                try:
                    current_unit.save()
                except Exception as e:
                    pprint(current_unit.to_dict())
                    print(e.message)
                finally:
                    current_event.status = 'complete'
                    current_event.save()

            # ensure competitors added to competition


            set_checksum("BELGIUM_IMPORT", "RESULTS", comp_num, results_chk_text)
            # mc.update_summary_stats()
            calculate_competition_summary(mc.id)
            if comp_num_wanted and comp_num==comp_num_wanted:
                break

    not_changed = not_changed.__self__
    unchanged_comps = len(not_changed)
    _showList(not_changed,'comps were unchanged')
    _showList(not_in_db,'comps were not in the db or events')

    if verbosity:
        print(('Processed %d comp_nums (%d changed, %d unchanged) in %.2f seconds' % (len(results_by_competition),
                    changed_comps, unchanged_comps, time.process_time()-clock_lsa)))
        # At end, some stats on what we did
        print(("Results processed: %d"%results_processed))
        print(("Failed event lookups: %d" % failed_event_lookups))
        print(("Failed athlete lookups: %d" % failed_athlete_lookups))
        print(("Failed discipline lookups: %d" % failed_discipline_lookups))

def load_competitions():
    # Return a mapping of the national ID to the competition
    comps = Competition.objects.filter(country_id="BEL")
    dikt = {}
    for comp in comps:
        if comp.national_id:
            dikt[comp.national_id.lower()] = comp
    return dikt

_eventmap = {}
def load_event_map():
    global _eventmap
    if _eventmap:
        return _eventmap
    else:
        _eventmap = {}
        for row in EVENTMAP:
            if len(row) > 1:
                evname = row[0]
                evcode = row[1]
                if not evcode:
                    continue
                _eventmap[evname] = evcode
        return _eventmap


def event_list_checksum(evlist):
    "Work out the checksum of list of events for this competition"
    lines = []
    for (event, categ) in evlist:
        lines.append(event + "," + categ)
    txt = "\n".join(lines)
    return hashlib.md5(asBytes(txt)).hexdigest()

def category_code(bel_cat):
    "Convert Belgian category like 'Masters Mannen' to something shorter"
    # Modified in Jan - they have 3 word ones like Alle Cat Mannen
    words = bel_cat.strip().split()
    return words[0][0:3].upper() + '-' + words[-1][0] if words else '??'


def load_season_athletes():
    "Return a map of (year, bib) to person"

    fed = Organisation.objects.get(country_id="BEL", code="KBAB")
    fedroles = Role.objects.filter(organisation=fed,role_type='REGISTERED').select_related('person')
    rmap = {}
    for r in fedroles:
        if r.identifier and r.bib:
            yyyy = r.date_to.year
            rmap[(yyyy, r.bib)] = r.person
    return rmap

def load_clubs():
    "Return a map of all clubs as dicts to set on the team property"

    clubs = Organisation.objects.filter(country_id="BEL", org_type="CLUB")
    cmap = {}
    for c in clubs:
        cmap[c.code] = dict(
                id=c.id,
                code=c.code,
                name=c.local_name()
            )
    return cmap


def category_suffix(bel_cat):
    """This works on a spelled out category in the Excel input like 'Masters M'
    but needs to be consistent with the belgian_category_suffix function in agegroups.py
    """
    cc = category_code(bel_cat)
    if cc[0:3] in ("SEN", "MAS"):
        l = 'S'
    elif cc[0:3] == 'SCH':
        l = 'L'
    elif cc[0:3] == 'CAD':
        l = 'K'
    else:
        l = cc[0]
    r = cc[-1]

    assert (l in 'SJLKMPB'), 'unexpected category in spreadsheet: %s, %s' % (bel_cat, cc)
    assert r in 'MV', 'unexpected category gender suffix in spreadsheet' + bel_cat

    return l + r


COUNTRYCODES = {
    "Macedonia": "MKD",
    "Ivory Coast": "CIV",
    "Roumenia": "ROU",
    "Syria": "SYR",
    "Guinee": "GNQ",
    "Korea": "KOR",
    "Erkend vluchteling": "BEL",
    "Surinam": "SUR",
    "England": "GBR",
    "Russian Federation": "RUS",
    "Serbia and Montenegro": "SRB",
}

__unknown_events__="""
  ["100m petra M", ],
  ["100m petra V", ],
  ["30 meter", ],
  ["40 meter Hindernissen", ],
  ["40m", ],
  ["discus seated 1kg M", ],
  ["discus seated 1kg V", ],
  ["kogel seated M", ],
  ["kogel seated V", ],
  ["Mixed Relay 4 x 400m", ],
  ["speer seated 600g M", ],
  ["speer seated 600g V", ],
  ["Stoten", ],
  ["Teamestafette", ],
  ["tienkamp (Jun)", ],
  ["tienkamp mannen", ],
  ["tienkamp sch", ],
  ["tienkamp vrouwen", ],
  ["vijfkamp (M50-55)", ],
  ["vijfkamp (M60-65)", ],
  ["vijfkamp (M70+)", ],
  ["vijfkamp M80+", ],
  ["vijfkamp W35", ],
  ["vijfkamp W50-55", ],
  ["vijfkamp W60-70", ],
  ["vijfkamp W70+", ],
  ["Vortexwerpen", ],
"""

EVENTMAP = json.loads("""
[
  ["10.000 meter","10000",null  ],
  ["1000m","1000",null  ],
  ["100m","100",null  ],
  ["100m H (0,762m)","100H",0.762  ],
  ["100m H (0,840m)","100H",0.84  ],
  ["100m H (0,914m)","100H",0.914  ],
  ["100m horden 0,838m M60-65", "100H", 0.84  ],
  ["100m horden 0,840", "100H", 0.84  ],
  ["100m horden 0,84m M60-65","100H", 0.84  ],
  ["100m petra M","",null  ],
  ["100m petra V","",null  ],
  ["100m wheelers M","100",null  ],
  ["100m wheelers V","100",null  ],
  ["10 km snelwandelen", "10KW", null],
  ["110m H (0,914m)","110H",0.914  ],
  ["110m H (0,990m)","110H",0.99  ],
  ["110m H (1,067m)","110H",1.067  ],
  ["1500m","1500",null  ],
  ["1500m Steeple (0,762m)","1500SC",null  ],
  ["150m","150",null  ],
  ["150m H (0,762m)","150H",0.762  ],
  ["150m horden 0,686", "150H", 0.686  ],
  ["2000m","2000",null  ],
  ["2000m Steeple (0,762m)","2000SC",0.762  ],
  ["2000m steeple 0,762m M60+","2000SC",0.762  ],
  ["2000m steeple 0,838m", "2000SC", 0.838 ],
  ["2000m steeple 0,914m","2000SC",0.914  ],
  ["200m","200",null  ],
  ["200m H (0,762m)","200H",0.762  ],
  ["200m horden 0,686m W70+", "200H", 0.686  ],
  ["200m wheelers M","200",null  ],
  ["200m wheelers V","200",null  ],
  ["20 km snelwandelen","20KW",null  ],
  ["2 mijl","2M",null  ],
  ["3000m","3000",null  ],
  ["3000m Steeple (0,762m)","3000SC",0.762  ],
  ["3000m Steeple (0,914m)","3000SC",0.914  ],
  ["300m","300",null  ],
  ["300m H (0,686m) - M70+ W60+","300H",0.686  ],
  ["300m H (0,762m)","300H",0.762  ],
  ["3 km snelwandelen","3KW",null  ],
  ["3x600m","3x600",null  ],
  ["400m","400",null  ],
  ["400m H (0,762m)","400H",null  ],
  ["400m H (0,840m)","400H",null  ],
  ["400m H (0,914m)","400H",null  ],
  ["400m horden 0,838m", "400H", 0.84  ],
  ["400m wheelers M","400",null  ],
  ["400m wheelers V","400",null  ],
  ["4x100m","4x100",null  ],
  ["4x1500m","4x1500",null  ],
  ["4x200m","4x200",null  ],
  ["4x400m","4x400",null  ],
  ["4x60m","4x60",null  ],
  ["4x800m","4x800",null  ],
  ["4x80m","4x80",null  ],
  ["5000m","5000",null  ],
  ["500 meter", "500", null],
  ["50m","50",null  ],
  ["50m H (0,762m) - Cad/W40-45","50H",0.762  ],
  ["50m H (0,840m) - M60-65","50H",0.84  ],
  ["5 km snelwandelen","5KW",null  ],
  ["600m","600",null  ],
  ["60m","60",null  ],
  ["60m H (0,686m) - M80+","60H",0.686  ],
  ["60m H (0,762m)","60H",0.762  ],
  ["60m H (0,840m)","60H",0.84  ],
  ["60m H (0,914m) - Cad/M50-55","60H",0.914  ],
  ["60m H (0,990m) - Jun/M35-45","60H",0.99  ],
  ["60m H (1,067m)","60H",1.067  ],
  ["60m horden 0,686", "60H", 0.686  ],
  ["60m horden 0,686m","60H",0.686  ],
  ["60m horden 0,686m W60+","60H",0.686  ],
  ["60m horden 0,838m", "60H", 0.84  ],
  ["60m horden 0,838m (cad)'", "60H", 0.84  ],
  ["800m","800",null  ],
  ["800m wheelers M","800",null  ],
  ["800m wheelers V","800",null  ],
  ["80m","80",null  ],
  ["80m H (0,686m) - W60+","80H",0.686  ],
  ["80m H (0,840m)","80H",0.84  ],
  ["80m horden 0,686m M80+","80H",0.686  ],
  ["80m horden 0,762m","80H",0.762  ],
  ["80m horden 0,762m M70-75","80H",0.762  ],
  ["80m horden 0,762m W50-55","80H",0.762  ],
  ["achtkamp","OCT",null  ],
  ["discus seated 1kg M","SDT1Kg",null  ],
  ["discus seated 1kg V","SDT1Kg",null  ],
  ["discuswerpen 0,750kg W75+","DT0.75k",0.75  ],
  ["discuswerpen (1,5kg)","DT1.5k",1.5  ],
  ["discuswerpen (1,75kg)","DT1.75k",1.75  ],
  ["discuswerpen (1kg)","DT1k",1 ],
  ["discuswerpen (2kg)","DT2k",2 ],
  ["discuswerpen (600g)","DT0.6k",0.6  ],
  ["discuswerpen (750g)","DT0.75k",0.75 ],
  ["gewichtw. (11,34kg) - M50-55","WT11.34k",11.34  ],
  ["gewichtw. (15,88kg)","WT15.88k",15.88  ],
  ["gewichtw. (4kg) - W75+","WT4k",4    ],
  ["gewichtw. (5,45kg) - M80+","WT5.45k",5.45  ],
  ["gewichtw. (5,45kg) - W60-70","WT5.45k",5.45 ],
  ["gewichtw. (7,26kg) - M70-75","WT7.26k",7.26 ],
  ["gewichtw. (7,26kg) - W50-55","WT7.26k",7.26 ],
  ["gewichtw. (9,08kg) - M60-65","WT9.08k",9.08 ],
  ["gewichtw. (9,08kg)","WT9.08k",9.08  ],
  ["hamer 6,000kg M50,M55","HT6k",6  ],
  ["hamerslingeren (2kg)","HT2k",2  ],
  ["hamerslingeren (3kg)","HT3k",3  ],
  ["hamerslingeren (4kg)","HT4k",4  ],
  ["hamerslingeren (5kg)","HT5k",5  ],
  ["hamerslingeren (6kg)","HT6k",6  ],
  ["hamerslingeren (7,260kg)","HT7.26k",7.26  ],
  ["hinkstapspringen","TJ",null  ],
  ["hockeybalwerpen","OT",null  ],
  ["hoogspringen","HJ",null  ],
  ["Kids Athletics","",null  ],
  ["kogel (3kg) M80+","SP3k",3  ],
  ["kogel seated M","SBT",null  ],
  ["kogel seated V","SBT",null  ],
  ["kogelstoten (1kg)","SP1k",1  ],
  ["kogelstoten (2kg)","SP2k",2  ],
  ["kogelstoten (3kg)","SP3k",3  ],
  ["kogelstoten (4kg)","SP4k",4  ],
  ["kogelstoten (5kg)","SP5k",5  ],
  ["kogelstoten (6kg)","SP6k",6  ],
  ["kogelstoten (7,260kg)","SP7k",7  ],
  ["Medley Aflossing (100-200-300-400)","4xRELAY",null  ],
  ["mijl","MILE",null  ],
  ["Olymp. Aflossing 800-400-200-200","4xRELAY",null  ],
  ["polsstokspringen","PV",null  ],
  ["speer seated 600g M","SJT600",null  ],
  ["speer seated 600g V","SJT600",null  ],
  ["speerwerpen (400g)","JT400",400  ],
  ["speerwerpen (500g)","JT500",500  ],
  ["speerwerpen (600g)","JT600",600  ],
  ["speerwerpen (700g)","JT700",700  ],
  ["speerwerpen (800g)","JT800",800  ],
  ["Tienkamp","DEC",null  ],
  ["uurloop", "1HR", null ],
  ["verspringen Ben","LJ",null  ],
  ["verspringen","LJ",null  ],
  ["vierkamp","QUAD",null  ],
  ["vijfkamp (cad)","PEN",null  ],
  ["vijfkamp (cad vr)","PEN",null  ],
  ["vijfkamp (jun)","PEN",null  ],
  ["vijfkamp (M35-45)","PEN",null  ],
  ["vijfkamp (min)","PEN",null  ],
  ["vijfkamp","PEN",null  ],
  ["vijfkamp (sch)","PEN",null  ],
  ["vijfkamp Sch-V-I","PEN",null  ],
  ["vijfkamp (sch vr)","PEN",null  ],
  ["vijfkamp (sen)","PEN",null  ],
  ["vijfkamp (W40-45)","PEN",null  ],
  ["werpvijfkamp M50-55","PEN",null  ],
  ["werpvijfkamp M60-65","PEN",null  ],
  ["werpvijfkamp M70-75","PEN",null  ],
  ["werpvijfkamp M80+","PEN",null  ],
  ["werpvijfkamp","PEN",null  ],
  ["werpvijfkamp V","PEN",null  ],
  ["werpvijfkamp W50-55","PEN",null  ],
  ["werpvijfkamp W60...70","PEN",null  ],
  ["werpvijfkamp W75+","PEN",null  ],
  ["zeskamp","HEX",null  ],
  ["zevenkamp","HEP",null  ],
  ["zevenkamp (jun)","HEP",null  ],
  ["zevenkamp Sch-M","HEP",null  ],
  ["zevenkamp Sch-V","HEP",null  ],
  ["zevenkamp (sen)","HEP",null  ],
  ["Zweedse Aflossing 400-300-200-100","4xRELAY",null  ]
]

""")
